
local MAJOR = "StarLibTimer-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LibTimer = LibStub:NewLibrary(MAJOR, MINOR)
if not LibTimer then return end
local error = LibStub("StarLibError-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

if not LibTimer.__index then
	LibTimer.__index = LibTimer
end


function stopTimers()
	local stop = true
	for i, v in ipairs(objects) do
		if v.active then
			stop = false
		end
	end
	
	if stop then
		frame:SetScript("OnUpdate", nil)
	end
end

function startTimers()
	local start = false
	for i, v in ipairs(objects) do
		if v.active then
			start = true
		end
	end
	
	if start then
		frame:SetScript("OnUpdate", update)
	end
end

-- @name StarLibTimer:New
-- @usage StarLibTimer:New(duration, repeating, callback, data)
-- @param duration How long the timer should be in milliseconds.
-- @param repeating Whether to repeat the timer or not
-- @param callback Function to call when the timer has expired
-- @param data Data to pass to the callback
-- @param errorLevel Error verbosity level
-- @param durationLimit How much to throttle the timer. Be nice. 10 milliseconds is default.
-- @return A new StarLibTimer object
function LibTimer:New(duration, repeating, callback, data, errorLevel, durationLimit)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.duration = duration / 1000
	obj.repeating = repeating
	obj.callback = callback
	obj.data = data
	obj.errorLevel = errorLevel
	obj.durationLimit = (durationLimit or 10) / 1000
	
	return obj	
	
end

-- @name StarLibTimer:Del
-- @usage StarLibTimer:Del([timer]) or object:Del()
-- @param timer An optional timer object
-- @return Nothing
function LibTimer:Del(timer)
	
	if timer and type(timer) == "table" then
		pool[timer] = true
		for i, v in ipairs(objects) do
			if v == timer then
				objectsDict[timer] = nil
				tremove(objects, i)
			end
		end
	else
		pool[self] = true
		for i, v in ipairs(objects) do
			if v == timer then
				objectsDict[timer] = nil
				tremove(objects, i)
			end
		end
	end
end

-- @name StarLibTimer:Start
-- @usage object:Start()
-- @return Nothing
function LibTimer:Start()
	self.active = true
	self.startTime = GetTime()	
	startTimers()
end

-- @name LibTimer:Stop
-- @usage object:Stop()
function LibTimer:Stop()
	self.active = false
	stopTimers()
end

-- @name LibTimer:TimeRemaining
-- @usage object:TimeRemaining()
-- @return The remaining duration
function LibTimer:TimeRemaining()
	if type(self.startTime) ~= "number" or not self.active then return 0 end
	
	local time = GetTime()
	local diff = time - self.startTime
	
	return time - diff
end

local function timerUpdate(self)	
	if self.timer < self.durationLimit then
		return
	end
	
	if self.timer > self.duration then
		self:Stop()
		if self.repeating then self:Start() end
		if self.callback then self.callback(self.data) end
		self.timer = 0
		return
	end
end

update = function(self, elapsed)
	for i, o in ipairs(objects) do
		if o.active then
			o.timer = (o.timer or 0) + elapsed
			timerUpdate(o)	
		end
	end
end

